par.ori <- par(no.readonly=TRUE)
## Preciso numrica
set.seed(1)

n <- 100
x1 <- sort(runif(n))
y <- 0.2*(x1-0.5) + (x1-0.5)^2 + rnorm(n, sd=0.1)
x1
y
## modelo de regresso polinomial:
reg1 <- lm(y ~ x1+I(x1^2))
reg1
## fazendo matricialmente...
X <- cbind(1, x1, I(x1^2))
X
## formas alternativas/equivalentes (e mais interessantes!) de montar a matriz X
##X <- model.matrix(reg) 
##X <- model.matrix(~ x1+I(x1^2)) 
beta.hat <- solve(t(X)%*%X, t(X)%*%y) 
beta.hat


## deslocando os valores de x
x2 <- x1 + 100
reg2 <- lm(y ~ x2+I(x2^2))
reg2
## mas fazendo matricialmente agora (ingenuamente...)
X <- model.matrix(~ x2+I(x2^2)) 
beta.hat <- solve(t(X)%*%X, t(X)%*%y) 
## ou de outra forma...
beta.hat <- solve(crossprod(X), crossprod(X,y))
## evitando o erro (util para manipulao de opes dentro de funes)
beta.hat <- try(solve(crossprod(X), crossprod(X,y)), silent=T)
beta.hat
if(class(beta.hat) == "try-error")
  beta.hat <- "SOLUES ALTERNATIVAS AQUI (VER MAIS ADIANTE)"

## mas mesmo lm() pode falhar ou produzir resultados estranhos! (mesmo que no aparente isto!!!)
x3 <- x1 + 1000
reg3 <- lm(y ~ x3 + I(x3^2))
reg3 

## e compare os modelos ajustados nos grficos
par(mfrow=c(1,3))
plot(x1, y)
#xvals <- seq(min(x1), max(x1), len=100)
#lines(xvals, predict(reg1, newdata=data.frame(x1 = xvals)))
lines(x1, fitted(reg1))
##
plot(x2, y)
#xvals <- seq(min(x2), max(x2), len=100)
#lines(xvals, predict(reg2, newdata=data.frame(x2 = xvals)))
lines(x1, fitted(reg1))
##
plot(x3, y)
#xvals <- seq(min(x3), max(x3), len=100)
#lines(xvals, predict(reg3, newdata=data.frame(x3 = xvals)))
lines(x3, fitted(reg3))

par(par.ori)

## Investigar tb o comportamento usando poly() para definir a regresso polinomial!!!


#####################################################################
## Eficincia em clculo matricial
n <- 1000
A <- matrix(runif(n^2), n, n)
B <- matrix(runif(n*n), n, n)
y <- runif(n)
system.time(A%*%B%*%y) 
system.time(A%*%(B%*%y))

## verificar nmero de "flops" envolvidos nas operaes!!

## Trao de matrix
n <- 5000
m <- 100
A <- matrix(runif(n*m), n, m)
B <- matrix(runif(n*m), m, n)
system.time(sum(diag(A%*%B))) # 2n^2m
system.time(sum(diag(B%*%A))) # 2nm^2 ( tr(AB) = tr(BA) )
system.time(sum(A*t(B)))      # 2nm   (tr(AB) = sum_ij Aij Bji

## Cholesky
## S = R'R ;   R  triangular superior

## simulao de NMV: y = mu + R'z
S <- matrix(c(5,-3, 2, -3,5,-3, 2,-3,5), 3, 3)
S
mu <- c(1, 0, -4)
R <- chol(S) 
R
t(R)
z <- rnorm(3) 
y <- mu + t(R)%*%z
y
y <- mu + crossprod(R,z)
y

## exemplo no modelo Geo

## verificar/ilustrar comportamento de mdias... varincias e correlaes!
z <- matrix(rnorm(30), nc=10) 
y <- mu + t(R)%*%z
## que  equivalente a:
y <- mu + crossprod(R, z)
round(y, dig=3)

## comparando parmetros verdadeiros e estimativas pelas simulaes
# mdias
mu
rowMeans(y)
# varincias
diag(S)
apply(y, 1, var)
# correlaes
Cor <- S/outer(diag(S), diag(S), function(x,y) sqrt(x)*sqrt(y))
Cor[lower.tri(Cor)]
cor.d <- cor(t(y))
cor.d[lower.tri(cor.d)]

## Uso no clculo de verossimilhana (modelo normal mv)
y <- 1:3
## l(\theta) = -(n/2) log(2*\pi) - (1/2) log(|\Sigma|) - (1/2) (y - F\beta)'\Sigma^{-1} (y - F\beta)

## clculo "ingnuo"
logLik <- -0.5*(length(y) * log(2*pi) + determinant(S, log=TRUE)$mod +
  crossprod((y - mu), solve(S, (y - mu))))
logLik

## e a verossimilhana pode ser calculada mais adequadamente por:
z <- forwardsolve(t(R), y-mu)
z
logLik <- -length(y)*log(2*pi)/2 - sum(log(diag(R))) - sum(z*z)/2
logLik
## notar que det(S) = (prod(R_ii))^2


## Autovalores e autovetores
set.seed(123)
x1 <- sort(runif(100))
x <- x1 + 100 
X <- model.matrix(~ x + I(x^2)) 
lambda <- eigen(t(X)%*%X)$values
lambda
lambda <- eigen(crossprod(X))$values
lambda
lambda[1]/lambda[3]

## Decomposio de valor singular
d <- svd(X)$d 
d
d[1]/d[3]


## Decomposio QR: X = QR,  R tring sup e Q'Q = Im (mas QQ' != I)
X.qr <- qr(X) 

## revisitar soluo do sistema de equaes normais
## e o clculo da funo de verossimilhana da normal multivariada utilizando:
#   (i) a decomposio spectral,
##  (ii) decomposio de valores singulares e
##  (iii) decomposio QR


######################################################
## Operaes de clculo no computador

## Um pouco mais sobre operaes numricas - erro de cancelamento
a <- 1e12
b <- 1e12 + pi
d <- b - a
d
pi
## ###
a <- 1e16
b <- 1e16 + pi
d <- b - a
d
pi
## erro: !!!! (veja o tamanho do erro relativo)
(d-pi)/pi 

## derivao automtica
dauto <- function(x, diff = c(1,1)) {
  ## cria classe dauto
  ## comprimento de diff[1]  o meso de grad
  grad <- rep(0, diff[1])
  if (diff[2]>0 && diff[2]<=diff[1]) grad[diff[2]] <- 1
  attr(x,"grad") <- grad
  class(x) <- "dauto"
  x
}
x1 <- dauto(1,c(3,1))
x1

## pode-se definir fcs matemticas e operadores para a classe "dauto" e propagar as derivadas
seno.dauto <- function(a) {
  grad.a <- attr(a,"grad")
  a <- as.numeric(a) ## para editar recursso infinita....
  d <- sin(a)
  attr(d, "grad") <- cos(a) * grad.a ## regra da cadeia...
  class(d) <- "dauto"
  d
}

sin(x1)
## resultados: seno de x1 e primeiro elemento de grad  a derivada com respeito a 

"*.ad" <- function(a,b) { 
  grad.a <- attr(a,"grad")
  grad.b <- attr(b,"grad")
  a <- as.numeric(a)
  b <- as.numeric(b)
  d <- a*b ## avaliando...
  attr(d,"grad") <- a * grad.b + b * grad.a ## regra da cadeia
  class(d) <- "dauto"
  d
}

## e pode-se desta forma contruir uma "biblioteca" de funes e operaes para a classe
## com ela pode-se obter as derivadas diretamento do cdigo
## Ex:
x1 <- 1
x2 <- 2
x3 <- pi/2
(x1*x2*sin(x3)+ exp(x1*x2))/x3
## e agora novamenteo com argumentos substituidos por objetos "dauto"
x1 <- dauto(1,c(3,1))
x2 <- dauto(2,c(3,2))
x3 <- dauto(pi/2,c(3,3))
(x1*x2*sin(x3)+ exp(x1*x2))/x3


########################################################################

# Integrao Numrica

## quadraturas -- funo escada
# ex: \int_0^1 exp(x) dx

## soluo analtica:
I <- exp(1) - 1
I

## soluo por integrate() (uma dimenso e limitada)
I.int <- integrate(function(x) exp(x), lower=0, upper=1)
I.int
c(I.int$val, I, (I-I.int$val)/I)

## approximao por ponto mdio
N <- 10
I.pm <- sum(exp((1:N-.5)/N))/N
I.pm
c(I.pm, I, (I-I.pm)/I)

## pode-se aumentar a acurcia...
N <- 100
I.pm <- sum(exp((1:N-.5)/N))/N
c(I.pm, I, (I-I.pm)/I)

N <- 1000
I.pm <- sum(exp((1:N -.5)/N))/N
c(I.pm, I, (I-I.pm)/I)

## quadratura gaussiana
require(statmod)
N <- 10
## pontos x_i de integrao e pesos w_i
qg <- gauss.quad(N) 
qg
## a funo  reescrita para aplicar a quadratura e no intervalo [-1,1]
## usando a relao \int^a_b f(x) dx = \int_{-1}^1 f(/frac{a+b+(b-a)t}{2}) dt
I.qg <- sum(qg$weights * exp((qg$nodes+1)/2))/2
c(I.qg,I,(I-I.qg)/I)

## reduzindo o nmero de pontos...
N <- 5
gq <- gauss.quad(N) 
I.qg <- sum(gq$weights*exp((gq$nodes+1)/2))/2
c(I.qg, I, (I-I.qg)/I)

N <- 3
qg <- gauss.quad(N) 
I.qg <- sum(qg$weights*exp((qg$nodes+1)/2))/2
c(I.qg, I, (I - I.qg)/I)

## examinar os argumentos da funo para outras opes e mtodos! 

## outras dadas opes so ainda dadas em gauss.quad.prob()
args(gauss.quad.prob)

## uniforme
qgu <- gauss.quad.prob(10, dist="uniform", l=0, u=1)
qgu
I.qgu <- sum(qgu$w * exp(qgu$n))
I.qgu
c(I.qgu, I, (I - I.qgu)/I)

## normal (no  muito natural para este exemplo, pode-se tentar
## diferentes opes para mu e sigma mas aproximao no funciona muito bem...)
## Tem-se aqui o problema do domnio de integrao estar restrito de em [0,1] 
## Esta opo  mais adequada para variveis como domnio de integrao entre (-Inf, Inf) 
## e requer que mu e sigma sejam informados
qgn <- gauss.quad.prob(10, dist="normal", mu=0.5, sigma=0.2)
qgn

## idem anterior para Gamma com parmetro de forma alpha e escala beta
## s que neste caso com intervalo (0, +Inf)
qgg <- gauss.quad.prob(10, dist="gamma", alpha=1, beta=0.5)
qgg
## e note que para alpha (shape) = 1 tem-se a exponencial
ff <- function(x) dgamma(x, sh=1, sc=1/5)
curve(ff, from=0, to=3)
#
dgamma(0:5, sh=1, rate=1/10)
dexp(0:5, rate=1/10)

## Outra alternativa para o intervalo (0, 1)  dado pela Beta
##
bb <- function(x, a, b) x^(a-1)*(1-x)^(b-1)/beta(a,b)
bb((0:10)/10, 2, 3)
dbeta((0:10)/10, 2, 3)
ff <- function(x) dbeta(x, 2, 2)
curve(ff, 0, 1)
ff <- function(x) dbeta(x, 0.5, 0.5)
curve(ff, 0, 1, add=T, lty=2)
ff <- function(x) dbeta(x, 1.5, 0.5)
curve(ff, 0, 1, add=T, lwd=2)
ff <- function(x) dbeta(x, 1, 1)
curve(ff, 0, 1, add=T, lty=2, lwd=2)

## mas note que a qualidade a aproximao depende da escolha dos parmetros
qgb <- gauss.quad.prob(10, dist="beta", a=2, b=2)
qgb
I.qgb <- sum(qgb$w * exp(qgb$n))
I.qgb
c(I.qgb, I, (I - I.qgb)/I)

qgb <- gauss.quad.prob(10, dist="beta", a=0.5, b=0.5)
qgb
I.qgb <- sum(qgb$w * exp(qgb$n))
I.qgb
c(I.qgb, I, (I - I.qgb)/I)

qgb <- gauss.quad.prob(10, dist="beta", a=1.5, b=0.5)
qgb
I.qgb <- sum(qgb$w * exp(qgb$n))
I.qgb
c(I.qgb, I, (I - I.qgb)/I)

qgb <- gauss.quad.prob(10, dist="beta", a=1, b=1)
qgb
I.qgb <- sum(qgb$w * exp(qgb$n))
I.qgb
c(I.qgb, I, (I - I.qgb)/I)

##
## Integrao Multidimensional...
##

## baseado na aplicao recursiva de regras de 1-D
## \int \int f(x,y) dx dy \approx \sum_j \sum_i w_j w_i f(x_j, y_i)

## funo para encontrar conjunto de ns e pesos d-dimensional,
malha <- function(nos, dim, pesos = 1/length(x)+x*0) {
  ## nos: ns (em 1-D) onde a funo deve ser avaliada
  ## dim: dimenso do espao de integrao
  ## pesos: pesos correspondentes aos ns 
  n <- length(nos)
  X <- W <- matrix(0, n^dim, dim)
  for (i in 1:dim) {
    X[,i] <- nos
    W[,i] <- pesos
    nos <- rep(nos, rep(n, length(nos)))
    pesos <- rep(pesos, rep(n, length(pesos)))
  }
  ## pesos so dados pelo produto das colunas de W
  pesos <- exp(rowSums(log(W))) 
  ## e cada linha de X fornece as coordenadas dos ns
  list(nodes=X, weights=pesos) 
}

## Ex: integrar f(x) = exp(\sum x_i), em [-1; 1]^d
fd <- function(x) {exp(rowSums(x))}

N <- 10; d <- 3
## nmero de avaliaes da funo:
N^d

## soluo analtica
I <- (exp(1)-exp(-1))^d

## Aproximao por ponto mdio
malha.pm <- malha((1:N-.5)/N*2-1, d, rep(2/N,N))
I.pm <- sum(malha.pm$weights * fd(malha.pm$nodes))
c(I.pm, I,(I-I.pm)/I)

## agora em 6-D 
N <- 10; d <- 6
## nmero de avaliaes da funo
N^d

I <- (exp(1)-exp(-1))^d
malha.pm <- malha((1:N-.5)/N*2-1, d, rep(2/N,N))
I.pm <- sum(malha.pm$wei * fd(malha.pm$nodes))
c(I.pm, I,(I-I.pm)/I)


## agora por quadratura gaussiana, aumenta (mas nem tanto!) o nmero de avaliaes da funo
N <- 5; d <- 2
N^d

I <- (exp(1)-exp(-1))^d
gq <- gauss.quad(N)
malha.qg <- malha(gq$nodes, d, gq$weights)
I.qg <- sum(malha.qg$wei * fd(malha.qg$nodes))
c(I.qg, I, (I-I.qg)/I)

## comparar o efeito do aumento da dimenso
N <- 10; d <- 6
N^d

I <- (exp(1)-exp(-1))^d
qg <- gauss.quad(N)
malha.gq <- malha(gq$nodes, d, gq$weights)
I.gq <- sum(malha.qg$wei * fd(malha.qg$nodes))
c(I.qg, I, (I-I.qg)/I)


## experimentar com outros nmeros de ns etc etc
                                                     

## Aproximando o integrando: abordagem por Aproximao de Laplace


## Integrao de Monte Carlo

## Monte Carlo estratificado

## Amostragem por importncia (Importance sampling)

## Amostragem por importncia via Laplace (Laplace importance sampling)


## Exemplo: Poisson com efeitos aleatrios
## Modelo:
##   y_i|b \sim P(\mu_i)
##   log(\mu_i) = \beta_0 \beta_1 x_i + b_1 z1_i + b2 z2_i
## em forma matricial:
##   log(\mu_i) = X_i \beta  + Z_i b
##   Z dada por z1 e z2, assumidas aqui serem variveis binrias (por ex, indicadoras de grupos)
##   b \sim N(0, V)  ; diag(V) = (\sigma^2_1, \sigma^2_1, \sigma^2_2, \sigma^2_2)
## ou em termos de matrix de preciso
##   b \sim N(0, V^{-1})  ; diag(V^{-1}) = (\psi1, \psi_1, \psi_2, \psi_2)

## simulao com \beta_0 = 0, \beta_1 = 3, \sigma_1^2 = 1 e  \sigma_2^2 = 1 

set.seed(0)
n <- 50
# valores da covarivel (parte fixa):
x <- runif(n)
X <- model.matrix(~x)
##
## valores das covariveis (parte aleatria)
z1 <- sample(c(0,1), n, replace=TRUE)
z2 <- sample(c(0,1), n, replace=TRUE)
Z <- cbind(z1,1-z1,z2,1-z2)
## simulando os efeitos aleatrios:
b <- rnorm(4) 
b
##
## simulando dados
eta <- X %*% c(0,3) + Z%*%b
mu <- exp(eta)
y <- rpois(length(mu), mu)
y
## os parmetros do modelo usados na simulao so portanto (0, 3, 1, 1)

## Note que neste modelo todos indivduos do mesmo grupo compartilham o mesmo valor
## (realizado) do efeito aleatrio
## em cada varivel z, os dois grupos possuem o memos valor do parmetro de varincia

## Vero:
## L(y) = \int f(y, b) db = \int f(y|b) f(b) db = \int exp{log{f(y|b) f(b)}}

## ef. aleatrios independentes ==> f(b) = \prod f(b_j)
## log(f(y, b)) = log(f(y|b) f(b))
##  \propto \sum_i \{y_i(X_i \beta + Z_i b)-exp(X_i \beta + Z_i b)-log(y!)\} - 0.5 \sum_j \psi_j b_j^2


## funo para avaliar log{f(y; b)} (log usado para evitar problemas numricos com zeros)
lf.yb0 <- function(y, b, theta, X, Z, lambda) {
  beta <- theta[1:ncol(X)]
  theta <- theta[-(1:ncol(X))]
  eta <- X%*%beta + Z%*%b
  mu <- exp(eta)
  lam <- lambda(theta, ncol(Z))
  sum(y*eta - mu -lfactorial(y)) - sum(lam*b^2)/2
}

## valor da densidade conjunta f(y,b) para os valores simulados
lf.yb0(y, b, c(1, 3, 1, 1), X=X, Z=Z, lambda=var.func)


## ***DEFINIR AGORA*** lf.yb() para aceitar b matricial (para calcular para vrios b de uma s vez)
## isto  utilneste exemplo para aproveitar clculos na integrao aproximada
lf.yb <- function(y, b, theta, X, Z, lambda) {
  beta <- theta[1:ncol(X)]
  theta <- theta[-(1:ncol(X))]
  ## replicando X%*%beta para somar aos vrios ef. aleatrios:
  ## eta <-  matrix(X%*%beta, nr=length(y), nc=ncol(as.matrix(b))) + Z%*%b 
  ##  ou ... transformando para vetor recicla X %*% beta
  eta <-  Z%*%b + drop(X%*%beta)  
  mu <- exp(eta)
  lam <- drop(lambda(theta, ncol(Z)))
  colSums(y*eta - mu -lfactorial(y)) - colSums(b^2 * lam)/2
}

## Exemplo: densidade conjunta para vrios (5 conjuntos) valores dos efeitos eleatrios
b.aleats <- cbind(b, matrix(rnorm(16), nc=4))
lf.yb(y, b=b.aleats, theta = c(1, 3, 1, 1), X=X, Z=Z, lambda=var.func)
## comparando e testando....
apply(b.aleats, 2,  function(bs) lf.yb0(y, bs, c(1, 3, 1, 1), X=X, Z=Z, lambda=var.func))


## elementos diagonais de V^{-1}
var.func <- function(theta, n.b) c(theta[1], theta[1], theta[2], theta[2])

## quadratura por ponto mdio

## Problema: a integral  definida em toda reta real.
## Soluo: como E(b) = 0 vamos restringir a integral em $ [-5, 5]^d $ 
theta <- c(0,3,1,1)

nm <- 10   # nmero de pontos
m.r <- 10  # amplitude: 5 -(-5)
## grid de pontos para b em 4-D onde a funo ser avaliada:
bm <- malha((1:nm-(nm+1)/2)*m.r/nm, 4, rep(m.r/nm,nm))

lf <- lf.yb(y,t(bm$nodes),theta, X, Z, var.func)
log(sum(bm$wei * exp(lf-max(lf)))) + max(lf)


## Note que usamos aqui (e nos resultados a seguir) que:
## $\log\left\{\sum_i w_i exp\{t_i\}\right\} = \log\left\[w_i exp\{t_i - max(t_i)\}\right\] + max(t_i)


## aumentando (dobrando no ex) nm pode-se aumentar a preciso
nm <- 20
bm <- malha((1:nm-(nm+1)/2)*m.r/nm,4,rep(m.r/nm,nm))

lf <- lf.yb(y,t(bm$nodes), theta, X, Z, var.func)
log(sum(bm$wei * exp(lf-max(lf)))) + max(lf)

## quadratura Gaussiana (potencialmente pode exigir um menor nmero de avaliaes da funo...)
nm <- 10
qg <- gauss.quad(nm)
qg
## rescalonando ao domnio
x <- qg$nodes*m.r/2
w <- qg$weights * m.r/2 
x
w

bm <- malha(x, 4, w)
lf <- lf.yb(y, t(bm$nodes), theta, X, Z, var.func)
log(sum(bm$wei * exp(lf-max(lf)))) + max(lf)

## muito longe de aproximao de ponto mdio!!
# experimente dobrar mn para 20
nm <- 20
gq <- gauss.quad(nm)
## rescalonando ao domnio:
w <- gq$weights*m.r/2 
x <- gq$nodes * m.r/2
bm <- malha(x,4,w)
log(sum(bm$w*exp(lf-max(lf)))) + max(lf)
## ainda no fica bom...

## PARA CONFERIR AQUI:
## verificar o reescalonamento ao domnio como visto anteriormente!! 

## a razo  que esta funo no  to suave...
## aproximao por ponto mdio parece melhor neste caso
## os resultados da quadratura Gaussiana pode ser melhorados com opes/configuraes adicionais
## no lugar de usar opes "default" que nem sempre funcionam bem....

## PARA FAZER: experimentar a quadratura de Gauss Hermite, adequada para integra;o em (-Inf, +Inf)
## \int e^(-x/2) f(x) dx = \sum_i w_i f(x_i)
## gauss.quad(10, kind="hermite")

## Aproximao de Laplace

## antes definir aqui as funes glf.yb() e hlf.yb()
## usar o clculo de gradiente conforme visto acima (ou entao alternativas tais como
## deriv(), deriv3() etc

## para clculo do gradiente e hessiano conforme exigido pela aproximao de Laplace

## O primeiro passo  a obteno de \hat{b}. Abaixo isto  feito pelo mtodo de Newton-Rapson
b.hat <- rep(0,4)
while(TRUE) {
  b0 <- b.hat
  g <- glf.yb(y, b.hat, theta, X, Z, var.func)
  H <- hlf.yb(y, b.hat, theta, X, Z, var.func)
  b.hat <- b.hat - solve(H,g)
  ## fixando critrio de parada/convergncia:
  if (sum(abs(b.hat-b0.hat))<1e-10*sum(abs(b.hat))) break
}
## ao final deste passo b.hat que comeou com 0's tem seus valores substituidos por $\hat{b}$


## verificar outras opes para se obter \hat{b}!!!

## aproximao de Laplace para f(y):
## f_y(y) = f(y, \hat{b}) 2\pi^{d/2} |H|^{-1/2}

## e portanto, para $log\{f(y)\}$ temos
lf.yb(y, b.hat, theta, X, Z, var.func) + 2*log(2*pi) - 0.5*determinant(H, log=TRUE)$modulus



# Integrao de Monte Carlo
## \hat{I}_{MC} \approx \frac{V(\Omega)}{N} \sum_{i=1}{N} f(x_i)

## nmero de pontos para aproximar a integral
N <- 100000
## sorteando os pontos de uma Uniforme em 4-D, i.e. [-5, 5]^4
##bm <- matrix(runif(N*4, min=-5, max=5), 4, N)
bm <- matrix((runif(N*4)-.5)*m.r, 4, N)
hist(bm)
range(bm)
## avaliando a funo nestes N pontos
lf <- lf.yb(y,bm,theta,X,Z,var.func)

## avaliando a integral pela aproximao de Monte Carlo
log(sum(exp(lf-max(lf)))) + max(lf)  + log(m.r^4) - log(N)


## De forma mais geral Integrao de  Monte Carlo Pode ser escrita por:
## $\hat{I}_{MC} = \int h(x) f(x) dx \approx (1/n) \sum_{i=1}^n h(x_i) $
## e por importance sampling
## $\hat{I}_{IS} = \int h(x) f(x) dx \approx (1/n) \sum_{i=1}^n h(x_i) \frac{f(x_i)}{g(x_i)} $


## Amostragem por importncia via Laplace

## g(x)  substituida por uma Normal MV (\hat{x}, H^{-1})
## H  o Hessiano de -log(g.f) em relao a x avaliado em \hat{x}

## Adicionalmente,  vantajoso trabalhar com vetores NMV padronizados
## lembrando que z \sim N(0, I) e x \ \hat{x} + R^{-1} z
## R  Choleski de H, R'R = H
## A integral re-expressa em termos da densidade de z (em lugar de g) fica:
## $\hat{I}_{LIS} = \int h(x) f(x) dx
##  \approx \frac{2\pi^{d/2}}{n|R|} \sum_{i=1}^n g(\hat{x} + R^{-1} z) f(\hat{x} + R^{-1} z)
##                                                                    exp\{z^{'}_i z_i/2\} $

N <- 10000
R <- chol(-H)
z <- matrix(rnorm(N*4), 4, N)
bm <- b + backsolve(R,z)
lf <- lf.yb(y, bm, theta, X, Z, var.func)
zz.2 <- colSums(z*z)/2
lfz <- lf + zz.2
log(sum(exp(lfz - max(lfz)))) + max(lfz) + 2 * log(2*pi) - sum(log(diag(R))) - log(N)

## em ambos casos melhorias poderiam ser conseguidas com estratgias de estratificao

